/*  WBACRAESGenerator.cpp -- generator of White-Box Attack Resistant AES implementation.
 
    This is part of WBACR_AES project, generator of White-Box Attack Resistant AES implementation.
    See http://www.svenda.com/petr/dpthesis/index.html for more information
    Copyright (C) 2004  Petr Svenda <petr@svenda.com>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Please, report any bugs to author <petr@svenda.com>
/**/

#include <memory.h>
#include <crtdbg.h>
#include <iostream.h>

#include "WBACRAESGenerator.h"
#include "AES\rijndael-alg-ref.h"
#include "AES\rijndael-api-ref.h"
#include "AES\boxes-ref.dat"
#include "..\Shared\status.h"

int CWBACRAESGenerator::Cipher128_GenerateCodingMap(WBACR_AES_CODING_MAP* pCodingMap, int* pCodingsCount, AESMATRIX_CODING inputBlockEncoding, AESMATRIX_CODING outputBlockEncoding) {
    /*
     *  CODING MAP GENERATION RULES (DURING CIPHER OPERATION): 
     *  1. WHEN TABLE LOOKUP TO TABLE CELL X IS PERFORMED FOR VALUE Y, THAN COPY OUTPUT CODING ID FROM Y TO INPUT CODING OF X
     *     AND GENERATE NEW CODING ID FOR OUTPUT CODING OF X (ONE FOR LOW AND SECOND FOR HIGH 4-BITS)
     *  2. WHEN VALUE OF VARIABLE V IS COPIED TO ANOTHER VARIABLE W, THAN SET INPUT AND OUTPUT CODING FROM V TO W. 
     *  Example: 
     *   original command: W = ((V = lookupTables[Y]));
     *   replace by      : lookupTables.coding.IC = Y.coding.OC; 
     *                     lookupTables.coding.OC.L = ++codingID;     // low and high 4-bits coding are different            
     *                     lookupTables.coding.OC.H = ++codingID;                 
     *                     V.coding = lookupTables.coding;
     *                     W.coding = V.coding  
     *
     *  
     *  INPUT/OUTPUT BLOCK ENCODING:
     *  1. WHEN REQUIRED, PASS 'inputBlockEncoding' OR 'outputBlockEncoding' STRUCTURES, WHICH WILL CONTAINS CODING ID 
     *     USED FOR INPUT/OUTPUT BLOCK ENCODING. 
     *  2. FOR INPUT BLOCK ENCODING, IT WILL BE CODING ID's FROM 1 TO 32 
     *  3. FOR OUTPUT BLOCK ENCODING, IT WILL BE LAST 32 CODING ID's
     *  4. GENERATE CODINGS CORRESPONDING TO THIS CODING ID's BY YOU OWN REQUIREMENT
     *  
     */
    int     status = 0; 
    int     col;
    int     row;
    int     i;
    BYTE    round;
    BYTE    xorCounter;
    DWORD   codingID = 0;
    CODING  strip[STATE_ROWS];
    CODING  XOR_RES1;
    CODING  XOR_RES2;
    CODING  XOR_INPUT;
    AESMATRIXMC_CODING stateMC;
    AESMATRIX_CODING   state;
    AESMATRIX_CODING   state2;

    
    if (inputBlockEncoding != NULL) {
        // INITIAL STATE HAS INPUT CODING
        for (row = 0; row < STATE_ROWS; row++) {  
            for (col = 0; col < STATE_COLS; col++) {
                state[row][col].OC.H = inputBlockEncoding[row][col].OC.H = ++codingID;
                state[row][col].OC.L = inputBlockEncoding[row][col].OC.L = ++codingID;
            }
        }
    }
    else {
        // INITIAL STATE HAS NO INPUT/OUTPUT CODING
        for (row = 0; row < STATE_ROWS; row++) {  
            for (col = 0; col < STATE_COLS; col++) {
                state[row][col].OC.H = 0;
                state[row][col].OC.L = 0;
            }
        }
    }

    //
    // ORDINARY 9 ROUNDS
    //
    for (round = 0; (round < N_ROUNDS - 1); round++) {

        //
        //  OBTAIN MC STRIPS (4 BYTES) FOR EACH CELL OF STATE
        //
        for (row = 0; row < STATE_ROWS; row++) {  
            for (col = 0; col < STATE_COLS; col++) {
                // CLEAR TEMP VALUE CODING
                ResetCoding(stateMC[row][col][0]);
                ResetCoding(stateMC[row][col][1]);
                ResetCoding(stateMC[row][col][2]);
                ResetCoding(stateMC[row][col][3]);
                
                //** stateMC[row][col][0] =  m_pTables->ROUND[round] [(col + shifts[row][0]) % 4] [row] [state[row][(col + shifts[row][0]) % 4]] [0];
                //** stateMC[row][col][1] =  m_pTables->ROUND[round] [(col + shifts[row][0]) % 4] [row] [state[row][(col + shifts[row][0]) % 4]] [1];
                //** stateMC[row][col][2] =  m_pTables->ROUND[round] [(col + shifts[row][0]) % 4] [row] [state[row][(col + shifts[row][0]) % 4]] [2];
                //** stateMC[row][col][3] =  m_pTables->ROUND[round] [(col + shifts[row][0]) % 4] [row] [state[row][(col + shifts[row][0]) % 4]] [3];
                AssignCoding(pCodingMap->ROUND[round][(col + shifts[row][0]) % 4][row][0].IC, state[row][(col + shifts[row][0]) % 4].OC);
                AssignCoding(pCodingMap->ROUND[round][(col + shifts[row][0]) % 4][row][0].OC.L, ++codingID);
                AssignCoding(pCodingMap->ROUND[round][(col + shifts[row][0]) % 4][row][0].OC.H, ++codingID);
                AssignCoding(pCodingMap->ROUND[round][(col + shifts[row][0]) % 4][row][1].IC, state[row][(col + shifts[row][0]) % 4].OC);
                AssignCoding(pCodingMap->ROUND[round][(col + shifts[row][0]) % 4][row][1].OC.L, ++codingID);
                AssignCoding(pCodingMap->ROUND[round][(col + shifts[row][0]) % 4][row][1].OC.H, ++codingID);
                AssignCoding(pCodingMap->ROUND[round][(col + shifts[row][0]) % 4][row][2].IC, state[row][(col + shifts[row][0]) % 4].OC);
                AssignCoding(pCodingMap->ROUND[round][(col + shifts[row][0]) % 4][row][2].OC.L, ++codingID);
                AssignCoding(pCodingMap->ROUND[round][(col + shifts[row][0]) % 4][row][2].OC.H, ++codingID);
                AssignCoding(pCodingMap->ROUND[round][(col + shifts[row][0]) % 4][row][3].IC, state[row][(col + shifts[row][0]) % 4].OC);
                AssignCoding(pCodingMap->ROUND[round][(col + shifts[row][0]) % 4][row][3].OC.L, ++codingID);
                AssignCoding(pCodingMap->ROUND[round][(col + shifts[row][0]) % 4][row][3].OC.H, ++codingID);

                AssignCoding(stateMC[row][col][0], pCodingMap->ROUND[round][(col + shifts[row][0]) % 4][row][0]);
                AssignCoding(stateMC[row][col][1], pCodingMap->ROUND[round][(col + shifts[row][0]) % 4][row][1]);
                AssignCoding(stateMC[row][col][2], pCodingMap->ROUND[round][(col + shifts[row][0]) % 4][row][2]);
                AssignCoding(stateMC[row][col][3], pCodingMap->ROUND[round][(col + shifts[row][0]) % 4][row][3]);
            }
        }

        //
        //  PERFORM XOR OF STRIPS
        //
        for (row = 0; row < STATE_ROWS; row++) {   
            for (col = 0; col < STATE_COLS; col++) {
                
                // CLEAR TEMP VALUE CODING
                ResetCoding(strip[0]);
                ResetCoding(strip[1]);
                ResetCoding(strip[2]);
                ResetCoding(strip[3]);

                //** strip[0] = stateMC[0][col][row];
                //** strip[1] = stateMC[1][col][row];
                //** strip[2] = stateMC[2][col][row];
                //** strip[3] = stateMC[3][col][row];
                AssignCoding(strip[0], stateMC[0][col][row]);
                AssignCoding(strip[1], stateMC[1][col][row]);
                AssignCoding(strip[2], stateMC[2][col][row]);
                AssignCoding(strip[3], stateMC[3][col][row]);
                
                //
                // XOR ADDITIONS OF PREPARED STRIP
                //
            
                // CLEAR TEMP VALUE CODING
                ResetCoding(XOR_RES1);
                ResetCoding(XOR_RES2);

                // XORING PER 4-BITES OVER STRIP
                //** XOR_RES1 = CWBACRAESGenerator::LOWER(CWBACRAESGenerator::GET_LOWER(strip[0]));
                AssignCoding(XOR_RES1.IC.L, strip[0].IC.L);
                AssignCoding(XOR_RES1.OC.L, strip[0].OC.L);

                //** XOR_RES2 = CWBACRAESGenerator::HIGHER(CWBACRAESGenerator::GET_HIGHER(strip[0]));
                AssignCoding(XOR_RES2.IC.H, strip[0].IC.H);
                AssignCoding(XOR_RES2.OC.H, strip[0].OC.H);

                for (i = 0; i < STATE_ROWS - 1; i++) {
                    // calculating apropriate xor counter for this xoring: 6 xors (== three 8 bits xor => six 4-bits xors) per strip (== row)
                    xorCounter = row * 6 + i * 2;       
                    // XORING LOW 4-BITES 

                    // CLEAR TEMP VALUE CODING
                    ResetCoding(XOR_INPUT);

                    //** XOR_INPUT = CWBACRAESGenerator::LOWER(CWBACRAESGenerator::GET_LOWER(XOR_RES1)) | CWBACRAESGenerator::HIGHER(CWBACRAESGenerator::GET_LOWER(strip[i + 1]));
                    AssignCoding(XOR_INPUT.IC.L, XOR_RES1.IC.L);
                    AssignCoding(XOR_INPUT.OC.L, XOR_RES1.OC.L);
                    AssignCoding(XOR_INPUT.IC.H, strip[i+1].IC.L);
                    AssignCoding(XOR_INPUT.OC.H, strip[i+1].OC.L);

                    //** XOR_RES1 = CWBACRAESGenerator::LOWER(m_pTables->XOR[round][col][xorCounter][XOR_INPUT]);
                    AssignCoding(pCodingMap->XOR[round][col][xorCounter].IC, XOR_INPUT.OC);
                    AssignCoding(pCodingMap->XOR[round][col][xorCounter].OC.L, ++codingID);
                    // NOTE: output of XOR is 4-bits so needs only low 4-bits encoding, UNUSED THIS: AssignCoding(pCodingMap->XOR[round][col][xorCounter].OC.H, ++codingID);

                    // CLEAR TEMP VALUE CODING
                    ResetCoding(XOR_RES1);

                    AssignCoding(XOR_RES1.IC.L, pCodingMap->XOR[round][col][xorCounter].IC.L);
                    AssignCoding(XOR_RES1.OC.L, pCodingMap->XOR[round][col][xorCounter].OC.L);

                    // XORING HIGH 4-BITES 

                    xorCounter = row * 6 + i * 2 + 1;   
                    
                    // CLEAR TEMP VALUE CODING
                    ResetCoding(XOR_INPUT);

                    //** XOR_INPUT = CWBACRAESGenerator::LOWER(CWBACRAESGenerator::GET_HIGHER(XOR_RES2)) | CWBACRAESGenerator::HIGHER(CWBACRAESGenerator::GET_HIGHER(strip[i + 1]));
                    AssignCoding(XOR_INPUT.IC.L, XOR_RES2.IC.H);
                    AssignCoding(XOR_INPUT.OC.L, XOR_RES2.OC.H);
                    AssignCoding(XOR_INPUT.IC.H, strip[i+1].IC.H);
                    AssignCoding(XOR_INPUT.OC.H, strip[i+1].OC.H);
                    
                    //** XOR_RES2 = CWBACRAESGenerator::HIGHER(m_pTables->XOR[round][col][xorCounter][XOR_INPUT]);
                    AssignCoding(pCodingMap->XOR[round][col][xorCounter].IC, XOR_INPUT.OC);
                    AssignCoding(pCodingMap->XOR[round][col][xorCounter].OC.L, ++codingID);
                    // NOTE: output of XOR is 4-bits so needs only low 4-bits encoding, UNUSED THIS: AssignCoding(pCodingMap->XOR[round][col][xorCounter].OC.H, ++codingID);

                    // CLEAR TEMP VALUE CODING
                    ResetCoding(XOR_RES2);

                    AssignCoding(XOR_RES2.IC.H, pCodingMap->XOR[round][col][xorCounter].IC.H);
                    // NOTE: XOR_RES2 is variable which contains higher 4-bits of target 8-bits strip xoring, 
                    // but is filled by output of XOR which gives result in lower 4-bits, so OC.H of XOR_RES2 is IC.L of XOR table lookup result
                    AssignCoding(XOR_RES2.OC.H, pCodingMap->XOR[round][col][xorCounter].OC.L);
                }

                // CLEAR TEMP VALUE CODING
                ResetCoding(state[row][col]);

                //** state[row][col] = CWBACRAESGenerator::LOWER(CWBACRAESGenerator::GET_LOWER(XOR_RES1));
                // state[row][col].IC USED NO MORE:
                //AssignCoding(state[row][col].IC.L, XOR_RES1.IC.L);
                //AssignCoding(state[row][col].IC.H, XOR_RES2.IC.H);
                
                //** state[row][col] |= CWBACRAESGenerator::HIGHER(CWBACRAESGenerator::GET_HIGHER(XOR_RES2));
                AssignCoding(state[row][col].OC.L, XOR_RES1.OC.L);
                AssignCoding(state[row][col].OC.H, XOR_RES2.OC.H);
            }    
        }
    }

    //
    // FINAL ROUND
    //
    memcpy(state2, state, sizeof(AESMATRIX_CODING));
    for (col = 0; col < STATE_COLS; col++) {
        for (row = 0; row < STATE_ROWS; row++) {
            
            //** state[row][col] = pCodingMap->FINALROUND[(col + shifts[row][0]) % 4][row][state2[row][(col + shifts[row][0]) % 4]];
            AssignCoding(pCodingMap->FINALROUND[(col + shifts[row][0]) % 4][row].IC, state2[row][(col + shifts[row][0]) % 4].OC);

            if (outputBlockEncoding) {
                // OUTPUT STATE HAS GIVEN OUTPUT CODING
                pCodingMap->FINALROUND[(col + shifts[row][0]) % 4][row].OC.L = outputBlockEncoding[row][col].OC.L = ++codingID;
                pCodingMap->FINALROUND[(col + shifts[row][0]) % 4][row].OC.H = outputBlockEncoding[row][col].OC.H = ++codingID;
            }
            else {
                // NO OUTPUT CODING OF FINAL ROUND
                AssignCoding(pCodingMap->FINALROUND[(col + shifts[row][0]) % 4][row].OC.L, NO_CODING);
                AssignCoding(pCodingMap->FINALROUND[(col + shifts[row][0]) % 4][row].OC.H, NO_CODING);
            }   
        }
    }


    *pCodingsCount = codingID;
    return status;
}

int CWBACRAESGenerator::Decipher128_GenerateCodingMap(WBACR_AES_CODING_MAP* pCodingMap, int* pCodingsCount, AESMATRIX_CODING inputBlockEncoding, AESMATRIX_CODING outputBlockEncoding) {
    /*
     *  SEE NOTES FROM Cipher128_GenerateCodingMap()
     *
     *
     */
    int     status = 0; 
    int     col;
    int     row;
    int     i;
    BYTE    round;
    BYTE    xorCounter;
    DWORD   codingID = 0;
    CODING  strip[STATE_ROWS];
    CODING  XOR_RES1;
    CODING  XOR_RES2;
    CODING  XOR_INPUT;
    AESMATRIXMC_CODING stateMC;
    AESMATRIX_CODING   state;
    AESMATRIX_CODING   state2;

    if (inputBlockEncoding != NULL) {
        // INITIAL STATE HAS INPUT CODING
        for (row = 0; row < STATE_ROWS; row++) {  
            for (col = 0; col < STATE_COLS; col++) {
                state[row][col].OC.H = inputBlockEncoding[row][col].OC.H = ++codingID;
                state[row][col].OC.L = inputBlockEncoding[row][col].OC.L = ++codingID;
            }
        }
    }
    else {
        // INITIAL STATE HAS NO INPUT/OUTPUT CODING
        for (row = 0; row < STATE_ROWS; row++) {  
            for (col = 0; col < STATE_COLS; col++) {
                state[row][col].OC.H = 0;
                state[row][col].OC.L = 0;
            }
        }
    }

    //
    // INVERSE OF SPECIAL 10. ROUND AND ORDINARY ROUNDS 9 to 2
    //
    for (round = N_ROUNDS-1; round > 0; round--) {

        //
        //  OBTAIN MC STRIPS
        //
        for (row = 0; row < STATE_ROWS; row++) {  
            for (col = 0; col < STATE_COLS; col++) {
                // CLEAR TEMP VALUE CODING
                ResetCoding(stateMC[row][col][0]);
                ResetCoding(stateMC[row][col][1]);
                ResetCoding(stateMC[row][col][2]);
                ResetCoding(stateMC[row][col][3]);

                //** stateMC[row][col][0] =  m_pTables->INVROUND[round] [(col + shifts[row][1]) % 4] [row] [state[row][(col + shifts[row][1]) % 4]] [0];
                //** stateMC[row][col][1] =  m_pTables->INVROUND[round] [(col + shifts[row][1]) % 4] [row] [state[row][(col + shifts[row][1]) % 4]] [1];
                //** stateMC[row][col][2] =  m_pTables->INVROUND[round] [(col + shifts[row][1]) % 4] [row] [state[row][(col + shifts[row][1]) % 4]] [2];
                //** stateMC[row][col][3] =  m_pTables->INVROUND[round] [(col + shifts[row][1]) % 4] [row] [state[row][(col + shifts[row][1]) % 4]] [3];
                AssignCoding(pCodingMap->INVROUND[round][(col + shifts[row][1]) % 4][row][0].IC, state[row][(col + shifts[row][1]) % 4].OC);
                AssignCoding(pCodingMap->INVROUND[round][(col + shifts[row][1]) % 4][row][0].OC.L, ++codingID);
                AssignCoding(pCodingMap->INVROUND[round][(col + shifts[row][1]) % 4][row][0].OC.H, ++codingID);
                AssignCoding(pCodingMap->INVROUND[round][(col + shifts[row][1]) % 4][row][1].IC, state[row][(col + shifts[row][1]) % 4].OC);
                AssignCoding(pCodingMap->INVROUND[round][(col + shifts[row][1]) % 4][row][1].OC.L, ++codingID);
                AssignCoding(pCodingMap->INVROUND[round][(col + shifts[row][1]) % 4][row][1].OC.H, ++codingID);
                AssignCoding(pCodingMap->INVROUND[round][(col + shifts[row][1]) % 4][row][2].IC, state[row][(col + shifts[row][1]) % 4].OC);
                AssignCoding(pCodingMap->INVROUND[round][(col + shifts[row][1]) % 4][row][2].OC.L, ++codingID);
                AssignCoding(pCodingMap->INVROUND[round][(col + shifts[row][1]) % 4][row][2].OC.H, ++codingID);
                AssignCoding(pCodingMap->INVROUND[round][(col + shifts[row][1]) % 4][row][3].IC, state[row][(col + shifts[row][1]) % 4].OC);
                AssignCoding(pCodingMap->INVROUND[round][(col + shifts[row][1]) % 4][row][3].OC.L, ++codingID);
                AssignCoding(pCodingMap->INVROUND[round][(col + shifts[row][1]) % 4][row][3].OC.H, ++codingID);

                AssignCoding(stateMC[row][col][0], pCodingMap->INVROUND[round][(col + shifts[row][1]) % 4][row][0]);
                AssignCoding(stateMC[row][col][1], pCodingMap->INVROUND[round][(col + shifts[row][1]) % 4][row][1]);
                AssignCoding(stateMC[row][col][2], pCodingMap->INVROUND[round][(col + shifts[row][1]) % 4][row][2]);
                AssignCoding(stateMC[row][col][3], pCodingMap->INVROUND[round][(col + shifts[row][1]) % 4][row][3]);
            }
        }

        //
        //  PERFORM XOR OF STRIPS
        //
        for (row = 0; row < STATE_ROWS; row++) {   
            for (col = 0; col < STATE_COLS; col++) {
                // CLEAR TEMP VALUE CODING
                ResetCoding(strip[0]);
                ResetCoding(strip[1]);
                ResetCoding(strip[2]);
                ResetCoding(strip[3]);

                //** strip[0] = stateMC[0][col][row];
                //** strip[1] = stateMC[1][col][row];
                //** strip[2] = stateMC[2][col][row];
                //** strip[3] = stateMC[3][col][row];
                AssignCoding(strip[0], stateMC[0][col][row]);
                AssignCoding(strip[1], stateMC[1][col][row]);
                AssignCoding(strip[2], stateMC[2][col][row]);
                AssignCoding(strip[3], stateMC[3][col][row]);

                //
                // XOR ADDITIONS OF PREPARED STRIP
                //
            
                // CLEAR TEMP VALUE CODING
                ResetCoding(XOR_RES1);
                ResetCoding(XOR_RES2);

                // XORING PER 4-BITES OVER STRIP
                //** XOR_RES1 = CWBACRAESGenerator::LOWER(CWBACRAESGenerator::GET_LOWER(strip[0]));
                AssignCoding(XOR_RES1.IC.L, strip[0].IC.L);
                AssignCoding(XOR_RES1.OC.L, strip[0].OC.L);

                //** XOR_RES2 = CWBACRAESGenerator::HIGHER(CWBACRAESGenerator::GET_HIGHER(strip[0]));
                AssignCoding(XOR_RES2.IC.H, strip[0].IC.H);
                AssignCoding(XOR_RES2.OC.H, strip[0].OC.H);

                for (i = 0; i < STATE_ROWS - 1; i++) {
                    // calculating apropriate xor counter for this xoring: 6 xors (== three 8 bits xor => six 4-bits xors) per strip (== row)
                    xorCounter = row * 6 + i * 2;       
                    // XORING LOW 4-BITES 

                    // CLEAR TEMP VALUE CODING
                    ResetCoding(XOR_INPUT);

                    //** XOR_INPUT = CWBACRAESGenerator::LOWER(CWBACRAESGenerator::GET_LOWER(XOR_RES1)) | CWBACRAESGenerator::HIGHER(CWBACRAESGenerator::GET_LOWER(strip[i + 1]));
                    AssignCoding(XOR_INPUT.IC.L, XOR_RES1.IC.L);
                    AssignCoding(XOR_INPUT.OC.L, XOR_RES1.OC.L);
                    AssignCoding(XOR_INPUT.IC.H, strip[i+1].IC.L);
                    AssignCoding(XOR_INPUT.OC.H, strip[i+1].OC.L);

                    //** XOR_RES1 = CWBACRAESGenerator::LOWER(m_pTables->INVXOR[round][col][xorCounter][XOR_INPUT]);
                    AssignCoding(pCodingMap->INVXOR[round][col][xorCounter].IC, XOR_INPUT.OC);
                    AssignCoding(pCodingMap->INVXOR[round][col][xorCounter].OC.L, ++codingID);
                    // NOTE: output of XOR is 4-bits so needs only low 4-bits encoding, UNUSED THIS: AssignCoding(pCodingMap->INVXOR[round][col][xorCounter].OC.H, ++codingID);

                    // CLEAR TEMP VALUE CODING
                    ResetCoding(XOR_RES1);

                    AssignCoding(XOR_RES1.IC.L, pCodingMap->INVXOR[round][col][xorCounter].IC.L);
                    AssignCoding(XOR_RES1.OC.L, pCodingMap->INVXOR[round][col][xorCounter].OC.L);

                    // XORING HIGH 4-BITES 

                    xorCounter = row * 6 + i * 2 + 1;   
                    
                    // CLEAR TEMP VALUE CODING
                    ResetCoding(XOR_INPUT);

                    //** XOR_INPUT = CWBACRAESGenerator::LOWER(CWBACRAESGenerator::GET_HIGHER(XOR_RES2)) | CWBACRAESGenerator::HIGHER(CWBACRAESGenerator::GET_HIGHER(strip[i + 1]));
                    AssignCoding(XOR_INPUT.IC.L, XOR_RES2.IC.H);
                    AssignCoding(XOR_INPUT.OC.L, XOR_RES2.OC.H);
                    AssignCoding(XOR_INPUT.IC.H, strip[i+1].IC.H);
                    AssignCoding(XOR_INPUT.OC.H, strip[i+1].OC.H);
                    
                    //** XOR_RES2 = CWBACRAESGenerator::HIGHER(m_pTables->INVXOR[round][col][xorCounter][XOR_INPUT]);
                    AssignCoding(pCodingMap->INVXOR[round][col][xorCounter].IC, XOR_INPUT.OC);
                    AssignCoding(pCodingMap->INVXOR[round][col][xorCounter].OC.L, ++codingID);
                    // NOTE: output of XOR is 4-bits so needs only low 4-bits encoding, UNUSED THIS: AssignCoding(pCodingMap->INVXOR[round][col][xorCounter].OC.H, ++codingID);

                    // CLEAR TEMP VALUE CODING
                    ResetCoding(XOR_RES2);

                    AssignCoding(XOR_RES2.IC.H, pCodingMap->INVXOR[round][col][xorCounter].IC.H);
                    // NOTE: XOR_RES2 is variable which contains higher 4-bits of target 8-bits strip xoring, 
                    // but is filled by output of XOR which gives result in lower 4-bits, so OC.H of XOR_RES2 is IC.L of XOR table lookup result
                    AssignCoding(XOR_RES2.OC.H, pCodingMap->INVXOR[round][col][xorCounter].OC.L);
                }

                // CLEAR TEMP VALUE CODING
                ResetCoding(state[row][col]);

                //** state[row][col] = CWBACRAESGenerator::LOWER(CWBACRAESGenerator::GET_LOWER(XOR_RES1));
                // state[row][col].IC USED NO MORE:
                //AssignCoding(state[row][col].IC.L, XOR_RES1.IC.L);
                //AssignCoding(state[row][col].IC.H, XOR_RES2.IC.H);
                
                //** state[row][col] |= CWBACRAESGenerator::HIGHER(CWBACRAESGenerator::GET_HIGHER(XOR_RES2));
                AssignCoding(state[row][col].OC.L, XOR_RES1.OC.L);
                AssignCoding(state[row][col].OC.H, XOR_RES2.OC.H);
            }    
        }
    }

    //
    // INVERSE SPECIAL FIRST ROUND
    //
    memcpy(state2, state, sizeof(AESMATRIX_CODING));
    for (col = 0; col < STATE_COLS; col++) {
        for (row = 0; row < STATE_ROWS; row++) {
            //** state[row][col] = pCodingMap->INVFIRSTROUND[(col + shifts[row][1]) % 4][row][state2[row][(col + shifts[row][1]) % 4]];
            AssignCoding(pCodingMap->INVFIRSTROUND[(col + shifts[row][1]) % 4][row].IC, state2[row][(col + shifts[row][1]) % 4].OC);
        
            if (outputBlockEncoding) {
                // OUTPUT STATE HAS GIVEN OUTPUT CODING
                pCodingMap->INVFIRSTROUND[(col + shifts[row][1]) % 4][row].OC.L = outputBlockEncoding[row][col].OC.L = ++codingID;
                pCodingMap->INVFIRSTROUND[(col + shifts[row][1]) % 4][row].OC.H = outputBlockEncoding[row][col].OC.H = ++codingID;
            }
            else {
                // NO OUTPUT CODING OF INVFIRSTROUND
                AssignCoding(pCodingMap->INVFIRSTROUND[(col + shifts[row][1]) % 4][row].OC.L, NO_CODING);
                AssignCoding(pCodingMap->INVFIRSTROUND[(col + shifts[row][1]) % 4][row].OC.H, NO_CODING);
            }
        }
    }

    *pCodingsCount = codingID;
    return status;
}

int CWBACRAESGenerator::Cipher128_GenerateTables(char* keyMaterial, WBACR_AES_TABLES* pTables) {
    int                     status = 0;
    BYTE                    i;
    BYTE                    r;
    BYTE                    j;
    int                     b;
    BYTE                    xorCount;
    BITS4                   first;
    BITS4                   second;
    BYTE                    decoded;
    BYTE                    mixed = 0;
    BYTE                    t;
    BYTE                    mc[STATE_ROWS];
    keyInstance             key;
    WBACR_AES_CODING_MAP    codingMap;
    int                     codingCount;


    // PREPARING KEY
    memset(&key, 0, sizeof(keyInstance));
    key.blockLen = 128;
    makeKey(&key, DIR_ENCRYPT, 128, keyMaterial);

    // GENERATE CODING MAP
    status = Cipher128_GenerateCodingMap(&codingMap, &codingCount);

    // PREPARING ALL 4-BITS ENCODING/DECODING BIJECTIONS
    BIJECT4X4*  pInputCodings = new BIJECT4X4[codingCount+1];
    BIJECT4X4*  pOutputCodings = new BIJECT4X4[codingCount+1];
    Generate4x4Bijections(pInputCodings, pOutputCodings, codingCount);

    // ADDROUNDKEY + SUBBYTES + SHIFT ROW TABLES + STRIPS FROM MIXING COLUMNS    (ROUND TABLE)
    // 1st ... 9th round

    memset(pTables, 0, sizeof(WBACR_AES_TABLES));
    for (r = 1; r < N_ROUNDS; r++) {                // ROUND
        for (j = 0; j < STATE_COLS; j++) {          // COLUMN
            for (i = 0; i < STATE_ROWS; i++) {      // ROW
                for (b = 0; b < 256; b++) {         // ALL POSSIBLE 8-BITS VALUE
                    // INPUT DATA DECODING
                    // NOTE: codingMap.ROUND[r-1][j][i][0].IC == codingMap.ROUND[r-1][j][i][1].IC == codingMap.ROUND[r-1][j][i][2].IC == codingMap.ROUND[r-1][j][i][3].IC
                    decoded = Decode8(b, codingMap.ROUND[r-1][j][i][0], pInputCodings);
                    
                    // AddRoundKey together with SubByte
                    t = S[(decoded ^ key.keySched[r-1][i][j])];
    
                    // Multiply part of MIXING COLUMNS 
                    MultiplyPartOfMixingColumns(t, i, mc);

                    pTables->ROUND[r-1][j][i][b][0] = Encode8(mc[0], codingMap.ROUND[r-1][j][i][0], pOutputCodings);
                    pTables->ROUND[r-1][j][i][b][1] = Encode8(mc[1], codingMap.ROUND[r-1][j][i][1], pOutputCodings);
                    pTables->ROUND[r-1][j][i][b][2] = Encode8(mc[2], codingMap.ROUND[r-1][j][i][2], pOutputCodings);
                    pTables->ROUND[r-1][j][i][b][3] = Encode8(mc[3], codingMap.ROUND[r-1][j][i][3], pOutputCodings);
                }
            }
        }
    }

    //  ADDITION PART OF MIXING COLUMNS (XOR TABLE)
    for (r = 1; r < N_ROUNDS; r++) {                                // ROUNDS
        for (j = 0; j < STATE_COLS; j++) {                          // COLUMNS  == 4 x DWORD
            for (xorCount = 0; xorCount < 24; xorCount++) {         // 3 x XOR FOR 8 4-BITS PARTS OF 32-BITS DWORD  
                for (b = 0; b < 256; b++) {                         // ALL POSSIBLE 8-BITS VALUE AS A INPUT OF XOR (2 x 4-BITS INPUT)
                    // INPUT DATA DECODING
                    decoded = Decode8(b, codingMap.XOR[r-1][j][xorCount], pInputCodings);
                    
                    // EXTRACT 4-BITS PARAMETRS
                    first = GET_LOWER(decoded);
                    second = GET_HIGHER(decoded);
                    
                    // PERFORM BIT XOR
                    t = first ^ second;     
                    
                    // OUTPUT DATA ENCODING
                    pTables->XOR[r-1][j][xorCount][b] = Encode4(t, codingMap.XOR[r-1][j][xorCount], pOutputCodings);      
                }
            }
        }
    }

    // FINAL 10. ROUND
    for (j = 0; j < STATE_COLS; j++) {           // COLUMN
        for (i = 0; i < STATE_ROWS; i++) {       // ROW
            for (b = 0; b < 256; b++) { // ALL POSSIBLE 8-BITS VALUE
                // Input data decoding
                decoded = Decode8(b, codingMap.FINALROUND[j][i], pInputCodings);
                
                // AddRoundKey together with SubByte and subkey denotes to shifted location i, j 
                t = S[(decoded ^ key.keySched[9][i][j])] ^ key.keySched[10][i][(j + shifts[i][1]) % 4];

                // OUTPUT DATA ENCODING INSERTED TO INVERSED SHIFTED POSITION (SHIFT ROW)
                pTables->FINALROUND[j][i][b] = Encode8(t, codingMap.FINALROUND[j][i], pOutputCodings);
            }
        }
    }

    delete[] pInputCodings;
    delete[] pOutputCodings;

    return status;
}


int CWBACRAESGenerator::Decipher128_GenerateTables(char* keyMaterial, WBACR_AES_TABLES* pTables) {
    int                     status = 0;
    BYTE                    i;
    BYTE                    r;
    BYTE                    j;
    int                     b;
    BYTE                    xorCount;
    BITS4                   first;
    BITS4                   second;
    BYTE                    decoded;
    BYTE                    mixed = 0;
    BYTE                    t;
    BYTE                    mc[STATE_ROWS];
    keyInstance             key;
    WBACR_AES_CODING_MAP    codingMap;
    int                     codingCount;

    // PREPARING KEY
    memset(&key, 0, sizeof(keyInstance));
    key.blockLen = 128;
    makeKey(&key, DIR_DECRYPT, 128, keyMaterial);


    // GENERATE CODING MAP
    status = Decipher128_GenerateCodingMap(&codingMap, &codingCount);

    // PREPARING ALL 4-BITS ENCODING/DECODING BIJECTIONS
    BIJECT4X4*  pInputCodings = new BIJECT4X4[codingCount+1];
    BIJECT4X4*  pOutputCodings = new BIJECT4X4[codingCount+1];
    Generate4x4Bijections(pInputCodings, pOutputCodings, codingCount);

    // INVERSE OF ROUNDS 10 TO 2
    for (r = N_ROUNDS; r > 1; r--) {                 // ROUND
        for (j = 0; j < STATE_COLS; j++) {           // COLUMN
            for (i = 0; i < STATE_ROWS; i++) {       // ROW
                for (b = 0; b < 256; b++) {          // ALL POSSIBLE 8-BITS VALUE
                    // Input data decoding
                    // NOTE: codingMap.ROUND[r-1][j][i][0].IC == codingMap.ROUND[r-1][j][i][1].IC == codingMap.ROUND[r-1][j][i][2].IC == codingMap.ROUND[r-1][j][i][3].IC
                    decoded = Decode8(b, codingMap.INVROUND[r-1][j][i][0], pInputCodings);
                    
                    if (r == N_ROUNDS) {
                        // SPECIAL INVERSE OF 10. ROUND
                        // AddRoundKey together with SubByte and subkey denotes to shifted location i, j 
                        t = Si[(decoded ^ key.keySched[r][i][j])] ^ key.keySched[r-1][i][(j + shifts[i][0]) % 4];
                    }
                    else {
                        // INVERSE OF ORDINARY ROUND
                        // SubByte
                        t = Si[decoded] ^ key.keySched[r-1][i][(j + shifts[i][0]) % 4];
                    }
    
                    // Multiply part of MIXING COLUMNS 
                    InvMultiplyPartOfMixingColumns(t, i, mc);


                    pTables->INVROUND[r-1][j][i][b][0] = Encode8(mc[0], codingMap.INVROUND[r-1][j][i][0], pOutputCodings);
                    pTables->INVROUND[r-1][j][i][b][1] = Encode8(mc[1], codingMap.INVROUND[r-1][j][i][1], pOutputCodings);
                    pTables->INVROUND[r-1][j][i][b][2] = Encode8(mc[2], codingMap.INVROUND[r-1][j][i][2], pOutputCodings);
                    pTables->INVROUND[r-1][j][i][b][3] = Encode8(mc[3], codingMap.INVROUND[r-1][j][i][3], pOutputCodings);
                }
            }
        }
    }

    // INVERSE OF ROUND 1
    for (j = 0; j < STATE_COLS; j++) {           // COLUMN
        for (i = 0; i < STATE_ROWS; i++) {       // ROW
            for (b = 0; b < 256; b++) {          // ALL POSSIBLE 8-BITS VALUE
                // Input data decoding
                decoded = Decode8(b, codingMap.INVFIRSTROUND[j][i], pInputCodings);
                
                // SubByte
                t = Si[decoded] ^ key.keySched[0][i][(j + shifts[i][0]) % 4];

                pTables->INVFIRSTROUND[j][i][b] = Encode8(t, codingMap.INVFIRSTROUND[j][i], pOutputCodings);
            }
        }
    }

    //  ADDITION PART OF MIXING COLUMNS (XOR TABLE)
    for (r = N_ROUNDS; r > 1; r--) {                                // ROUNDS
        for (j = 0; j < STATE_COLS; j++) {                          // COLUMNS  == 4 x DWORD
            for (xorCount = 0; xorCount < 24; xorCount++) {         // 3 x XOR FOR 8 4-BITS PARTS OF 32-BITS DWORD  
                for (b = 0; b < 256; b++) {                         // ALL POSSIBLE 8-BITS VALUE AS A INPUT OF XOR (2 x 4-BITS INPUT)
                    // INPUT DATA DECODING
                    decoded = Decode8(b, codingMap.INVXOR[r-1][j][xorCount], pInputCodings);
                    
                    // EXTRACT 4-BITS PARAMETRS
                    first = GET_LOWER(decoded);
                    second = GET_HIGHER(decoded);
                    
                    // PERFORM BIT XOR
                    t = first ^ second;     
                    
                    // OUTPUT DATA ENCODING
                    pTables->INVXOR[r-1][j][xorCount][b] = Encode4(t, codingMap.INVXOR[r-1][j][xorCount], pOutputCodings);      
                }
            }
        }
    }

    return status;
}

int CWBACRAESGenerator::Cipher128_SaveTables(WBACRAESHeaderFiles files, WBACR_AES_TABLES* pTables) {
    int     status = STAT_OK;
    int     pos;
    int     pos2;
    CFile   file;
    CFile   outFile;
    CString roundTableFileString;
    CString xorTableFileString;
    CString finalRoundTableFileString;
    CString values;
    CString value;
    CString round;
    CString help1;
    CString arrayName;
    CString oldArrayName;
    
    //
    // pTables->ROUND
    //
    if (status == STAT_OK) {
        roundTableFileString.Empty();
        values.Empty();

        if (file.Open(files.roundTablesTpl, CFile::modeRead)) {
            file.ReadHuge(roundTableFileString.GetBuffer(file.GetLength() + 1), file.GetLength());
            roundTableFileString.ReleaseBuffer();
         
            file.Close();
        }
        else status = -1;
    }
    if (status == STAT_OK) {
        cout << "--- Saving " << files.roundTables << " header file ";
        cout.flush();     
   
        // SEARCH FOR STR_ROUNDTABLES 
        if ((pos = roundTableFileString.Find(STR_ROUNDTABLES)) != -1) {
            pos2 = pos + strlen(STR_ROUNDTABLES);
            // GET USER DATA FROM #roundTables#USER_DATA#
            if ((pos2 = roundTableFileString.Find('#', pos2)) != -1) {
                // REPLACE '#roundTables#USER_DATA#' BY 'roundTablesUSER_DATA' 
                oldArrayName = arrayName = roundTableFileString.Mid(pos, (pos2 - pos + 1));
                arrayName.Replace("#", "");    
                roundTableFileString.Replace(oldArrayName, arrayName);

                pos = pos + arrayName.GetLength();

                if ((status = GenerateRoundsValuesString(pTables->ROUND, &values)) == STAT_OK) {
                   roundTableFileString.Insert(pos, (LPCTSTR) values);
                }                 
            }
            else status = -1;
        }
        else status = -1;
    }
    if (status == STAT_OK) {
        if (outFile.Open(files.roundTables, CFile::modeWrite | CFile::modeCreate)) {
            outFile.WriteHuge((LPCTSTR) roundTableFileString, roundTableFileString.GetLength()-1);
            outFile.Close();    
    
            cout << " success\n"; 
        }
        else {
            status = -1; 
            cout << " fail\n"; 
        }
        cout.flush();     
    }

    //
    // pTables->XOR
    //
    if (status == STAT_OK) {
        xorTableFileString.Empty();
        values.Empty();
        if (file.Open(files.xorTablesTpl, CFile::modeRead)) {
            file.ReadHuge(xorTableFileString.GetBuffer(file.GetLength() + 1), file.GetLength());
            xorTableFileString.ReleaseBuffer();
         
            file.Close();
        }
    }

    if (status == STAT_OK) {
        cout << "--- Saving " << files.xorTables << " header file ";
        cout.flush();     
           
        // SEARCH FOR STR_XORTABLES 
        if ((pos = xorTableFileString.Find(STR_XORTABLES)) != -1) {
            pos2 = pos + strlen(STR_XORTABLES);

            // GET USER DATA FROM #keyword#USER_DATA#
            if ((pos2 = xorTableFileString.Find('#', pos2)) != -1) {
                // REPLACE '#keyword#USER_DATA#' BY 'keywordUSER_DATA' 
                oldArrayName = arrayName = xorTableFileString.Mid(pos, (pos2 - pos + 1));
                arrayName.Replace("#", "");    
                xorTableFileString.Replace(oldArrayName, arrayName);

                pos = pos + arrayName.GetLength();

                if ((status = GenerateXorValuesString(pTables->XOR, &values)) == STAT_OK) {
                    xorTableFileString.Insert(pos, (LPCTSTR) values);
                }                 
            }
        }
        else status = -1;
    }

    if (status == STAT_OK) {
        if (outFile.Open(files.xorTables, CFile::modeWrite | CFile::modeCreate)) {
            outFile.WriteHuge((LPCTSTR) xorTableFileString, xorTableFileString.GetLength()-1);
            outFile.Close();     

            cout << " success\n"; 
        }
        else {
            status = -1; 
            cout << " fail\n"; 
        }
        cout.flush();     
    }

    //
    // pTables->FINALROUND
    //
    if (status == STAT_OK) {
        finalRoundTableFileString.Empty();
        values.Empty();
        if (file.Open(files.finalRoundTableTpl, CFile::modeRead)) {
            file.ReadHuge(finalRoundTableFileString.GetBuffer(file.GetLength() + 1), file.GetLength());
            finalRoundTableFileString.ReleaseBuffer();
         
            file.Close();
        }
        else status = -1;
    }
    if (status == STAT_OK) {
        cout << "--- Saving " << files.finalRoundTable << " header file ";
        cout.flush();     
           
        // SEARCH FOR STR_FINALROUNDTABLE 
        if ((pos = finalRoundTableFileString.Find(STR_FINALROUNDTABLE)) != -1) {
            pos2 = pos + strlen(STR_FINALROUNDTABLE);

            // GET USER DATA FROM #keyword#USER_DATA#
            if ((pos2 = finalRoundTableFileString.Find('#', pos2)) != -1) {
                // REPLACE '#keyword#USER_DATA#' BY 'keywordUSER_DATA' 
                oldArrayName = arrayName = finalRoundTableFileString.Mid(pos, (pos2 - pos + 1));
                arrayName.Replace("#", "");    
                finalRoundTableFileString.Replace(oldArrayName, arrayName);

                pos = pos + arrayName.GetLength();

                if ((status = GenerateFirstOrFinalValuesString(pTables->FINALROUND, &values)) == STAT_OK) {
                    finalRoundTableFileString.Insert(pos, (LPCTSTR) values);
                }                 
            }
        }
        else status = -1;
    }
    if (status == STAT_OK) {
        if (outFile.Open(files.finalRoundTable, CFile::modeWrite | CFile::modeCreate)) {
            outFile.WriteHuge((LPCTSTR) finalRoundTableFileString, finalRoundTableFileString.GetLength()-1);
            outFile.Close();    
            
            cout << " success\n"; 
        }
        else {
            status = -1; 
            cout << " fail\n"; 
        }
        cout.flush();     
    }

    return status;
}     

int CWBACRAESGenerator::Decipher128_SaveTables(WBACRAESHeaderFiles files, WBACR_AES_TABLES* pTables) {
    int     status = 0;
    int     pos;
    int     pos2;
    CFile   file;
    CFile   outFile;
    CString invRoundTableFileString;
    CString invXorTableFileString;
    CString invFirstRoundTableFileString;
    CString values;
    CString value;
    CString round;
    CString help1;
    CString arrayName;
    CString oldArrayName;
    
    //
    // pTables->INVROUND
    //
    if (status == STAT_OK) {
        invRoundTableFileString.Empty();
        values.Empty();
        if (file.Open(files.invRoundTablesTpl, CFile::modeRead)) {
            file.ReadHuge(invRoundTableFileString.GetBuffer(file.GetLength() + 1), file.GetLength());
            invRoundTableFileString.ReleaseBuffer();
         
            file.Close();
        }
        else status = -1; 
    }
    if (status == STAT_OK) {
        cout << "--- Saving " << files.invRoundTables << " header file ";
        cout.flush();     

        // SEARCH FOR STR_INVROUNDTABLES 
        if ((pos = invRoundTableFileString.Find(STR_INVROUNDTABLES)) != -1) {
            pos2 = pos + strlen(STR_INVROUNDTABLES);
            // GET USER DATA FROM #keyword#USER_DATA#
            if ((pos2 = invRoundTableFileString.Find('#', pos2)) != -1) {
                // REPLACE '#keyword#USER_DATA#' BY 'keywordUSER_DATA' 
                oldArrayName = arrayName = invRoundTableFileString.Mid(pos, (pos2 - pos + 1));
                arrayName.Replace("#", "");    
                invRoundTableFileString.Replace(oldArrayName, arrayName);

                pos = pos + arrayName.GetLength();

                if ((status = GenerateRoundsValuesString(pTables->INVROUND, &values)) == STAT_OK) {
                   invRoundTableFileString.Insert(pos, (LPCTSTR) values);
                }                 
            }
            else status = -1;
        }
        else status = -1;
    }

    if (status == STAT_OK) {
        if (outFile.Open(files.invRoundTables, CFile::modeWrite | CFile::modeCreate)) {
            outFile.WriteHuge((LPCTSTR) invRoundTableFileString, invRoundTableFileString.GetLength()-1);
            outFile.Close();    
            
            cout << " success\n"; 
        }
        else {
            status = -1; 
            cout << " fail\n"; 
        }
        cout.flush();     
    }

    //
    // pTables->INVXOR
    //
    if (status == STAT_OK) {
        invXorTableFileString.Empty();
        values.Empty();
        if (file.Open(files.invXorTablesTpl, CFile::modeRead)) {
            file.ReadHuge(invXorTableFileString.GetBuffer(file.GetLength() + 1), file.GetLength());
            invXorTableFileString.ReleaseBuffer();
         
            file.Close();
        }
        else status = -1;
    }

    if (status == STAT_OK) {
        cout << "--- Saving " << files.invXorTables << " header file ";
        cout.flush();     
           
        // SEARCH FOR STR_INVXORTABLES 
        if ((pos = invXorTableFileString.Find(STR_INVXORTABLES)) != -1) {
            pos2 = pos + strlen(STR_INVXORTABLES);

            // GET USER DATA FROM #keyword#USER_DATA#
            if ((pos2 = invXorTableFileString.Find('#', pos2)) != -1) {
                // REPLACE '#keyword#USER_DATA#' BY 'keywordUSER_DATA' 
                oldArrayName = arrayName = invXorTableFileString.Mid(pos, (pos2 - pos + 1));
                arrayName.Replace("#", "");    
                invXorTableFileString.Replace(oldArrayName, arrayName);

                pos = pos + arrayName.GetLength();

                if ((status = GenerateXorValuesString(pTables->INVXOR, &values)) == STAT_OK) {
                    invXorTableFileString.Insert(pos, (LPCTSTR) values);
                }                 
            }
            else status = -1;
        }
        else status = -1;
    }
    if (status == STAT_OK) {
        if (outFile.Open(files.invXorTables, CFile::modeWrite | CFile::modeCreate)) {
            outFile.WriteHuge((LPCTSTR) invXorTableFileString, invXorTableFileString.GetLength()-1);
            outFile.Close();     

            cout << " success\n"; 
        }
        else {
            status = -1; 
            cout << " fail\n"; 
        }
        cout.flush();     
    }

    //
    // pTables->INVFIRSTROUND
    //
    if (status == STAT_OK) {
        invFirstRoundTableFileString.Empty();
        values.Empty();
        if (file.Open(files.invFirstRoundTableTpl, CFile::modeRead)) {
            file.ReadHuge(invFirstRoundTableFileString.GetBuffer(file.GetLength() + 1), file.GetLength());
            invFirstRoundTableFileString.ReleaseBuffer();
         
            file.Close();
        }
        else status = -1;
    }
    if (status == STAT_OK) {
        cout << "--- Saving " << files.invFirstRoundTable << " header file ";
        cout.flush();     
           
        // SEARCH FOR STR_INVFIRSTROUNDTABLE 
        if ((pos = invFirstRoundTableFileString.Find(STR_INVFIRSTROUNDTABLE)) != -1) {
            pos2 = pos + strlen(STR_INVFIRSTROUNDTABLE);

            // GET USER DATA FROM #keyword#USER_DATA#
            if ((pos2 = invFirstRoundTableFileString.Find('#', pos2)) != -1) {
                // REPLACE '#keyword#USER_DATA#' BY 'keywordUSER_DATA' 
                oldArrayName = arrayName = invFirstRoundTableFileString.Mid(pos, (pos2 - pos + 1));
                arrayName.Replace("#", "");    
                invFirstRoundTableFileString.Replace(oldArrayName, arrayName);

                pos = pos + arrayName.GetLength();

                if ((status = GenerateFirstOrFinalValuesString(pTables->INVFIRSTROUND, &values)) == STAT_OK) {
                    invFirstRoundTableFileString.Insert(pos, (LPCTSTR) values);
                }                 
            }
            else status = -1;
        }
        else status = -1;
    }
    if (status == STAT_OK) {
        if (outFile.Open(files.invFirstRoundTable, CFile::modeWrite | CFile::modeCreate)) {
            outFile.WriteHuge((LPCTSTR) invFirstRoundTableFileString, invFirstRoundTableFileString.GetLength()-1);
            outFile.Close();    
            
            cout << " success\n"; 
        }
        else {
            status = -1; 
            cout << " fail\n"; 
        }
        cout.flush();     
    }

    return status;
}     

int CWBACRAESGenerator::Coding_GenerateCBCCompliant(CODING8X8_TABLE* pWBACRAESOutputCoding, CODING8X8_TABLE* pWBACRAESInputCoding, CODING8X8_TABLE* pDataCoding) {
    int                 status = STAT_OK;
    int                 i;
    int                 j;
    CODING8X8_TABLE_EX  inputCoding;
    CODING8X8_TABLE_EX  outputCoding;
    CODING8X8_TABLE_EX  dataCoding;
    DWORD               SIZE = 256;
    int                 x;
    int                 y;
    int                 A;
    int                 Q;
    list<int>           freeFList;

    
    //  F ... pWBACRAESOutputCoding
    //  G ... pDataCoding
    //  H ... pWBACRAESInputCoding

    //  REQUESTED BEHAVIOUR: F(x) XOR G(y) = H(x XOR y)                (V1)
    
    inputCoding.clear();
    outputCoding.clear();
    dataCoding.clear();
    freeFList.clear();

    InitRandomGenerator();

    //  FILL LIST OF FREE (NON-USED) FUNCTION VALUES FOR F (F IS BIJECTION)     
    for (i = 0; i < SIZE; i++) freeFList.push_back(i);

    //
    //  F(0) CHOOSED RANDOMLY 
    //
    if (status == STAT_OK) {
        status = GetRandomFromList(&freeFList, &outputCoding.coding[0], TRUE); 
        outputCoding.invCoding[outputCoding.coding[0]] = 0;
    }
    //
    //  H(0) CHOOSED RANDOMLY 
    //
    if (status == STAT_OK) {
        status = GetRandomFromInterval(SIZE, &inputCoding.coding[0]); 
        inputCoding.invCoding[inputCoding.coding[0]] = 0;
    }

    //
    //  G(0) CALC. FROM F(0), H(0) AND V1  
    //
    if (status == STAT_OK) {
        dataCoding.coding[0] = inputCoding.coding[0] ^ outputCoding.coding[0];
        dataCoding.invCoding[dataCoding.coding[0]] = 0;
    }

    //
    //  CHOOSE ALL F(s), WHERE s = 2 ^ n, n < {0, ..., log2(SIZE) - 1} 
    //  CALC. G(s) FROM F(s) & H(0) USING F(s) XOR G(s) == H(s XOR s) == H(0)
    //  CALC. H(s) FROM F(s) & G(0) USING F(s) XOR G(0) == H(s)
    //
    for (i = 1; i < SIZE;) {
        // F(s)
        GetRandomFromList(&freeFList, &outputCoding.coding[i], TRUE);  
        outputCoding.invCoding[outputCoding.coding[i]] = i;

        // G(s)
        dataCoding.coding[i] = outputCoding.coding[i] ^ inputCoding.coding[0];
        dataCoding.invCoding[dataCoding.coding[i]] = i;
        
        // H(s)
        inputCoding.coding[i] = outputCoding.coding[i] ^ dataCoding.coding[0];
        inputCoding.invCoding[inputCoding.coding[i]] = i;
        
        i *= 2;     
    }

    
    for (i = 1; i < SIZE; i++) {
        Q = inputCoding.coding[i];
        for (j = i + 1; j < SIZE; j++) {
            if (outputCoding.coding[j] != UNASSIGNED_CODING) {
                A = Q ^ outputCoding.coding[j];

                if (dataCoding.coding[i ^ j] != UNASSIGNED_CODING) {
                    // CHECK ALREADY ASSIGNED CODING, MUST BE SAME
                    _ASSERT(A == dataCoding.coding[i ^ j]);
                } 
                else {
                    // CALC G(i ^ j)
                    dataCoding.coding[i ^ j] = A;
                    dataCoding.invCoding[A] = i ^ j;
                    
                    // CALC F(i ^ j) USING F(i ^ j) XOR G(i^ j) == H(0)
                    outputCoding.coding[i ^ j] = dataCoding.coding[i ^ j] ^ inputCoding.coding[0];
                    outputCoding.invCoding[outputCoding.coding[i ^ j]] = i ^ j;
                    
                    // CALC. H(i ^ j) USING F(0) XOR G(i ^ j) == H(i ^ j)
                    inputCoding.coding[i ^ j] = outputCoding.coding[0] ^ dataCoding.coding[i ^ j];
                    inputCoding.invCoding[inputCoding.coding[i ^ j]] = i ^ j;
                }
            }
            else {
                // DO NOTHING
            }
        }           
    }

    // TESTS FOR NON-ASSIGNED CODINGS
    if (status == STAT_OK) {
        for (x = 0; x < SIZE; x++) {
            _ASSERT(outputCoding.coding[x] != UNASSIGNED_CODING);
            _ASSERT(inputCoding.coding[x] != UNASSIGNED_CODING);
            _ASSERT(dataCoding.coding[x] != UNASSIGNED_CODING);
        }            
    }

    // TEST ALL POSSIBLE VALUES INPUT VALUES (8-bits INPUT)
    if (status == STAT_OK) {
        for (x = 0; x < SIZE; x++) {
            for (y = 0; y < SIZE; y++) {
                _ASSERT((outputCoding.coding[x] ^ dataCoding.coding[y]) == inputCoding.coding[(x ^ y)]);
            }
        }            
    }


    //
    // FILL GENERATED CODINGS INTO RETURN STRUCTURES 
    //
    if (status == STAT_OK) {
        for (x = 0; x < SIZE; x++) {
            pWBACRAESOutputCoding->coding[x] = (BYTE) outputCoding.coding[x];
            pWBACRAESOutputCoding->invCoding[x] = (BYTE) outputCoding.invCoding[x];

            pWBACRAESInputCoding->coding[x] = (BYTE) inputCoding.coding[x];
            pWBACRAESInputCoding->invCoding[x] = (BYTE) inputCoding.invCoding[x];

            pDataCoding->coding[x] = (BYTE) dataCoding.coding[x];
            pDataCoding->invCoding[x] = (BYTE) dataCoding.invCoding[x];
        }            
    }

    return status;
}         

    //
    //  PRIVATE FUNCTIONS
    //

void CWBACRAESGenerator::AssignCoding(CODING& target, CODING source) {
    AssignCoding(target.OC, source.OC);
    AssignCoding(target.IC, source.IC);
}

void CWBACRAESGenerator::AssignCoding(HIGHLOW& target, HIGHLOW source) {
    AssignCoding(target.H, source.H);
    AssignCoding(target.L, source.L);
}
 
void CWBACRAESGenerator::AssignCoding(DWORD& target, DWORD source) {
    if (target == UNASSIGNED_CODING) {
        // NO CODING ASSIGNED YET
        _ASSERT(source != UNASSIGNED_CODING);
        target = source;
    }
    else {
        // CODING ALREADY ASSIGNED, SOURCE MUST BE SAME
        _ASSERT(target == source);    
    }
} 

void CWBACRAESGenerator::ResetCoding(CODING& target) {
    ResetCoding(target.IC);
    ResetCoding(target.OC);
}

void CWBACRAESGenerator::ResetCoding(HIGHLOW& target) {
    ResetCoding(target.H);
    ResetCoding(target.L);
}
 
void CWBACRAESGenerator::ResetCoding(DWORD& target) {
    target = UNASSIGNED_CODING;
} 


int CWBACRAESGenerator::Generate4x4Bijections(BIJECT4X4* pInputCodings, BIJECT4X4* pOutputCodings, DWORD codingCount) {
    int         status = STAT_OK;

    InitRandomGenerator();
    for (DWORD i = 0; i <= codingCount; i++) {
        if ((status = Generate4X4Bijection(pInputCodings[i], pOutputCodings[i])) != STAT_OK) break;
    }
    
    return status;    
}

BYTE CWBACRAESGenerator::Decode8(BYTE encoded, CODING coding, BIJECT4X4* pInputCodings) {
    _ASSERT(coding.IC.L != UNASSIGNED_CODING);
    _ASSERT(coding.IC.H != UNASSIGNED_CODING);

    BITS4           outLow = 0;
    BITS4           outHigh = 0;
    BYTE            out = 0;

    //
    //  DECODE GIVEN plain VALUE BY APROPRIATE BIJECTION CODING
    //
    if (coding.IC.L == 0) {
        outLow = CWBACRAESGenerator::GET_LOWER(encoded);
    }
    else {
        outLow = pInputCodings[coding.IC.L][CWBACRAESGenerator::GET_LOWER(encoded)];
    }

    if (coding.IC.H == 0) {
        outHigh = CWBACRAESGenerator::GET_HIGHER(encoded);
    }
    else {
        outHigh = pInputCodings[coding.IC.H][CWBACRAESGenerator::GET_HIGHER(encoded)];
    }

    out = outLow;
    out |= CWBACRAESGenerator::HIGHER(outHigh);
    
    return out;
}


BITS4 CWBACRAESGenerator::Encode4(BITS4 plain, CODING coding, BIJECT4X4* pOutputCodings) {
    _ASSERT(coding.OC.L != UNASSIGNED_CODING);
    
    BITS4           out = 0;

    //
    //  ENCODE GIVEN plain VALUE BY APROPRIATE BIJECTION CODING
    //
    if (coding.OC.L == 0) {
        out = CWBACRAESGenerator::GET_LOWER(plain);
    }
    else {
        out = pOutputCodings[coding.OC.L][CWBACRAESGenerator::GET_LOWER(plain)];
    }
    
    return out;
}

BYTE CWBACRAESGenerator::Encode8(BYTE plain, CODING coding, BIJECT4X4* pOutputCodings) {
    _ASSERT(coding.OC.L != UNASSIGNED_CODING);
    _ASSERT(coding.OC.H != UNASSIGNED_CODING);

    BITS4           outLow = 0;
    BITS4           outHigh = 0;
    BYTE            out = 0;

    //
    //  ENCODE GIVEN plain VALUE BY APROPRIATE BIJECTION CODING
    //
    if (coding.OC.L == 0) {
        outLow = CWBACRAESGenerator::GET_LOWER(plain);
    }
    else {
        outLow = pOutputCodings[coding.OC.L][CWBACRAESGenerator::GET_LOWER(plain)];
    }

    if (coding.OC.H == 0) {
        outHigh = CWBACRAESGenerator::GET_HIGHER(plain);
    }
    else {
        outHigh = pOutputCodings[coding.OC.H][CWBACRAESGenerator::GET_HIGHER(plain)];
    }

    out = outLow;
    out |= CWBACRAESGenerator::HIGHER(outHigh);
    
    return out;
}

int CWBACRAESGenerator::MultiplyPartOfMixingColumns(BYTE b, int row, BYTE* mc) {
    int     i;

    for (i = 0; i < 4; i++) {
        mc[i] = mul(MC[i][row], b);
    }

    return STAT_OK;    
}

int CWBACRAESGenerator::InvMultiplyPartOfMixingColumns(BYTE b, int row, BYTE* mc) {
    int     i;

    for (i = 0; i < 4; i++) {
        mc[i] = mul(MCi[i][row], b);
    }

    return STAT_OK;    
}

DWORD CWBACRAESGenerator::MatrixMultiply(BYTE b, DWORD dw) {
    int     i;
    DWORD   result = 0;
    BYTE    help = 0;


    for (i = 0; i < 4; i++) {
        ((BYTE*) &result)[i] = ((BYTE*) &dw)[i] * b;
    }

    return result;
}

BYTE CWBACRAESGenerator::mul(BYTE a, BYTE b) {
   // multiply two elements of GF(2^m)
   // needed for MixColumn and InvMixColumn
   
	if (a && b) return Alogtable[(Logtable[a] + Logtable[b])%255];
	else return 0;
}

int CWBACRAESGenerator::Generate4X4Bijection(BIJECT4X4 biject, BIJECT4X4 invBiject) {
    int         status = STAT_OK;
    list<BYTE>  freeList;
    list<BYTE>::iterator  iter;
    BYTE        j;
    BYTE        i;
    BYTE        target;
    BYTE        pos;

    // 
    //  RANDOM 4x4 BIJECTION
    //
    
    // INIT RND GENERATOR
    InitRandomGenerator();

    for (j = 0; j < 16; j++) freeList.push_back(j);    

    // FOR EACH POSSIBLE VALUE FROM 0 TO 15
    for (j = 0; j < 16; j++) {
        GetRandomFromInterval(freeList.size(), &pos);    
        
        // CHOOSE RANDOMLY ONE NUMBER FROM UNUSED TARGETS
        iter = freeList.begin();
        for (i = 0; i < pos; i++) iter++;

        // ASSIGN CHOOSEN TARGET AND REMOVE FROM FREE LIST
        target = *iter;
        biject[j] = target;
        invBiject[target] = j;
        freeList.erase(iter);    
    }

    return status;
}

int CWBACRAESGenerator::Generate8X8Bijection(BIJECT8X8 biject, BIJECT8X8 invBiject) {
    int         status = STAT_OK;
    list<BYTE>  freeList;
    list<BYTE>::iterator  iter;
    int         j;
    int         i;
    BYTE        target;
    BYTE        pos;


    // 
    //  RANDOM 8x8 BIJECTION
    //

    // INIT RND GENERATOR
    InitRandomGenerator();

    for (j = 0; j < 256; j++) freeList.push_back(j);    

    // FOR EACH POSSIBLE VALUE FROM 0 TO 255
    for (j = 0; j < 256; j++) {
        // CHOOSE RANDOMLY ONE NUMBER FROM UNUSED TARGETS
        GetRandomFromInterval(freeList.size(), &pos);    
        iter = freeList.begin();
        for (i = 0; i < pos; i++) iter++;

        // ASSIGN CHOOSEN TARGET AND REMOVE FROM FREE LIST
        target = *iter;
        biject[j] = target;
        invBiject[target] = j;
        freeList.erase(iter);    
    }

    return status;
}

int CWBACRAESGenerator::GenerateRoundsValuesString(ROUND_TABLES tables, CString* valueString) {
    CString values;
    CString value;
    CString round;

    values = " = {";
    // ROUND
    for (int i = 0; i < N_ROUNDS; i++) {
        if (i > 0) values += ", ";
        values += "\n{";
        for (int j = 0; j < STATE_COLS; j++) {
            if (j > 0) values += ", ";
            values += "\n{";
            for (int k = 0; k < STATE_ROWS; k++) {
                if (k > 0) values += ", ";
                values += "\n{";
            
                // ROUND_TABLE
                round.Empty();
                for (int m = 0; m < 256; m++) {
                    if (m > 0) round += ", ";
                    if ((m % 2) == 0) round += "\n{";
                    else round += "{";
                    // MCSTRIP
                    for (int n = 0; n < 4; n++) {
                        if (n > 0) round += ", ";
                        value.Format("0x%x", tables[i][j][k][m][n]);
                        round += value;
                    }
                    round += "}";
                }
                round += "\n}";

                values += round;
            }
            values += "\n}";
        }
        values += "\n}";

        cout << ".";
        cout.flush();
    }
    values += "\n}";       
    
    *valueString = values;

    return STAT_OK;    
}

int CWBACRAESGenerator::GenerateXorValuesString(XOR_TABLES tables, CString* valueString) {
    CString values;
    CString value;
    CString help1;
    CString round;

    values = " = {";
    // ROUND
    for (int i = 0; i < N_ROUNDS; i++) {
        if (i > 0) values += ", ";
        values += "\n{";

        for (int j = 0; j < STATE_COLS; j++) {
            if (j > 0) values += ", ";
            values += "\n{";

            help1.Empty();
            for (int k = 0; k < 24; k++) {
                if (k > 0) help1 += ", ";
                help1 += "\n{\n";

                // XOR_TABLE
                round.Empty();
                for (int m = 0; m < 256; m++) {
                    if (m > 0) {
                        round += ", ";
                        if ((m % 16) == 0) round += "\n";
                    }

                    value.Format("0x%x", tables[i][j][k][m]);
                    round += value;
                }
                round += "\n}";

                help1 += round;
            }
            values += help1;

            values += "\n}";
        }
        values += "\n}";
        cout << ".";
        cout.flush();
    }
    values += "\n}\n";   
    
    *valueString = values;

    return STAT_OK;    
}

int CWBACRAESGenerator::GenerateFirstOrFinalValuesString(SPECIAL_TABLE table, CString* valueString) {
    CString values;
    CString value;
    CString round;

    values = " = {";

    for (int j = 0; j < STATE_COLS; j++) {
        if (j > 0) values += ", ";
        values += "\n{";
        for (int k = 0; k < STATE_ROWS; k++) {
            if (k > 0) values += ", ";
            values += "\n{";
        
            round.Empty();
            for (int m = 0; m < 256; m++) {
                if (m > 0) {
                    round += ", ";
                    if ((m % 16) == 0) round += "\n";
                }

                value.Format("0x%x", table[j][k][m]);
                round += value;
            }
            round += "\n}";

            values += round;
        }
        values += "\n}";
    }
    values += "\n}";           

    *valueString = values;

    return STAT_OK;    
}

int CWBACRAESGenerator::GetRandomFromList(list<int>* pList, int* pRandom, BOOL erase) {
    int     status = STAT_OK;
    int     pos;
    int     i;
    list<int>::iterator iter;

    status = GetRandomFromInterval(pList->size(), &pos);
    
    iter = pList->begin(); 
    for (i = 0; i < pos; i++) iter++;
    
    *pRandom = *iter;

    if (erase) pList->erase(iter);
    
    return status;
}

int CWBACRAESGenerator::GetRandomFromInterval(int highBound, int *pRandom) {
    int     status = STAT_OK;
    
    //
    //  TODO: IF BETTER (PSEUDO) RANDOM GENERATOR IS AVAILABLE, JUST REPLACE rand() 
    //  rand() IS USED NOW
    //

    // GENERATE RANDOM NUMBER    
    if (pRandom) *pRandom = rand() % highBound;

    return status;
}

int CWBACRAESGenerator::GetRandomFromInterval(BYTE highBound, BYTE *pRandom) {
    int     status = STAT_OK;
    int     random;

    if ((status = GetRandomFromInterval(highBound, &random)) == STAT_OK) {
        *pRandom = (BYTE) random;
    }
    
    return status;
}

int CWBACRAESGenerator::InitRandomGenerator() {
    // INIT RANDOM NUMBER GENERATOR
    srand((unsigned int) time(NULL));

    return STAT_OK;
}
